from abc import abstractmethod
from NGram.SimpleSmoothing import SimpleSmoothing


class TrainedSmoothing(SimpleSmoothing):

    @abstractmethod
    def learnParameters(self, corpus: list, N: int):
        pass

    def newLowerBound(self, current: float, currentLowerBound: float,
                      currentUpperBound: float, numberOfParts: int) -> float:
        """
        Calculates new lower bound.

        PARAMETERS
        ----------
        current : float
            current value.
        currentLowerBound : float
            current lower bound
        currentUpperBound : float
            current upper bound
        numberOfParts : int
            number of parts between lower and upper bound.

        RETURNS
        -------
        float
            new lower bound
        """
        if current != currentLowerBound:
            return current - (currentUpperBound - currentLowerBound) / numberOfParts
        else:
            return current / numberOfParts

    def newUpperBound(self, current: float, currentLowerBound: float,
                      currentUpperBound: float, numberOfParts: int) -> float:
        """
        Calculates new upper bound.

        PARAMETERS
        ----------
        current : float
            current value.
        currentLowerBound : float
            current lower bound
        currentUpperBound : float
            current upper bound
        numberOfParts : int
            number of parts between lower and upper bound.

        RETURNS
        -------
        float
            new upper bound
        """
        if current != currentUpperBound:
            return current + (currentUpperBound - currentLowerBound) / numberOfParts
        else:
            return current * numberOfParts

    def train(self, corpus: list, nGram):
        """
        Wrapper function to learn parameters of the smoothing method and set the N-gram probabilities.

        PARAMETERS
        ----------
        corpus : list
            Train corpus used to optimize parameters of the smoothing method.
        nGram : NGram
            N-Gram for which the probabilities will be set.
        """
        self.learnParameters(corpus, nGram.getN())
        self.setProbabilitiesGeneral(nGram)



class CorpusTest:

    def readCorpus(self, fileName: str) -> list:
        corpus = []
        inputFile = open(fileName, "r")
        lines = inputFile.readlines()
        for line in lines:
            words = line.split(" ")
            corpus.append(words)
        inputFile.close()
        return corpus



import unittest

from NGram.LaplaceSmoothing import LaplaceSmoothing
from test.SimpleSmoothingTest import SimpleSmoothingTest


class LaplaceSmoothingTest(SimpleSmoothingTest, unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        simpleSmoothing = LaplaceSmoothing()
        self.simpleUniGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.simpleBiGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.simpleTriGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexUniGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexBiGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexTriGram.calculateNGramProbabilitiesSimple(simpleSmoothing)

    def test_PerplexitySimple(self):
        self.assertAlmostEqual(12.809502, self.simpleUniGram.getPerplexity(self.simpleCorpus), 4)
        self.assertAlmostEqual(6.914532, self.simpleBiGram.getPerplexity(self.simpleCorpus), 4)
        self.assertAlmostEqual(7.694528, self.simpleTriGram.getPerplexity(self.simpleCorpus), 4)

    def test_PerplexityComplex(self):
        self.assertAlmostEqual(4085.763010, self.complexUniGram.getPerplexity(self.testCorpus), 4)
        self.assertAlmostEqual(24763.660225, self.complexBiGram.getPerplexity(self.testCorpus), 4)
        self.assertAlmostEqual(49579.187475, self.complexTriGram.getPerplexity(self.testCorpus), 4)

    def test_CalculateNGramProbabilitiesSimple(self):
        self.assertEqual((5 + 1) / (35 + self.simpleUniGram.vocabularySize() + 1), self.simpleUniGram.getProbability("<s>"))
        self.assertEqual((0 + 1) / (35 + self.simpleUniGram.vocabularySize() + 1), self.simpleUniGram.getProbability("mahmut"))
        self.assertEqual((1 + 1) / (35 + self.simpleUniGram.vocabularySize() + 1), self.simpleUniGram.getProbability("kitabı"))
        self.assertEqual((4 + 1) / (5 + self.simpleBiGram.vocabularySize() + 1), self.simpleBiGram.getProbability("<s>", "ali"))
        self.assertEqual((0 + 1) / (2 + self.simpleBiGram.vocabularySize() + 1), self.simpleBiGram.getProbability("ayşe", "ali"))
        self.assertEqual(1 / (self.simpleBiGram.vocabularySize() + 1), self.simpleBiGram.getProbability("mahmut", "ali"))
        self.assertEqual((2 + 1) / (4 + self.simpleBiGram.vocabularySize() + 1), self.simpleBiGram.getProbability("at", "mehmet"))
        self.assertEqual((1 + 1) / (4.0 + self.simpleTriGram.vocabularySize() + 1), self.simpleTriGram.getProbability("<s>", "ali", "top"))
        self.assertEqual((0 + 1) / (1.0 + self.simpleTriGram.vocabularySize() + 1), self.simpleTriGram.getProbability("ayşe", "kitabı", "at"))
        self.assertEqual(1 / (self.simpleTriGram.vocabularySize() + 1), self.simpleTriGram.getProbability("ayşe", "topu", "at"))
        self.assertEqual(1 / (self.simpleTriGram.vocabularySize() + 1), self.simpleTriGram.getProbability("mahmut", "evde", "kal"))
        self.assertEqual((2 + 1) / (3.0 + self.simpleTriGram.vocabularySize() + 1), self.simpleTriGram.getProbability("ali", "topu", "at"))

    def test_CalculateNGramProbabilitiesComplex(self):
        self.assertEqual((20000 + 1) / (376019.0 + self.complexUniGram.vocabularySize() + 1), self.complexUniGram.getProbability("<s>"))
        self.assertEqual((50 + 1) / (376019.0 + self.complexUniGram.vocabularySize() + 1), self.complexUniGram.getProbability("atatürk"))
        self.assertEqual((11 + 1) / (20000.0 + self.complexBiGram.vocabularySize() + 1), self.complexBiGram.getProbability("<s>", "mustafa"))
        self.assertEqual((3 + 1) / (138.0 + self.complexBiGram.vocabularySize() + 1), self.complexBiGram.getProbability("mustafa", "kemal"))
        self.assertEqual((1 + 1) / (11.0 + self.complexTriGram.vocabularySize() + 1), self.complexTriGram.getProbability("<s>", "mustafa", "kemal"))
        self.assertEqual((1 + 1) / (3.0 + self.complexTriGram.vocabularySize() + 1), self.complexTriGram.getProbability("mustafa", "kemal", "atatürk"))


if __name__ == '__main__':
    unittest.main()



import unittest

from NGram.GoodTuringSmoothing import GoodTuringSmoothing
from test.SimpleSmoothingTest import SimpleSmoothingTest


class GoodTuringSmoothingTest(SimpleSmoothingTest, unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        simpleSmoothing = GoodTuringSmoothing()
        self.simpleUniGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.simpleBiGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.simpleTriGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexUniGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexBiGram.calculateNGramProbabilitiesSimple(simpleSmoothing)
        self.complexTriGram.calculateNGramProbabilitiesSimple(simpleSmoothing)

    def test_PerplexitySimple(self):
        self.assertAlmostEqual(14.500734, self.simpleUniGram.getPerplexity(self.simpleCorpus), 4)
        self.assertAlmostEqual(2.762526, self.simpleBiGram.getPerplexity(self.simpleCorpus), 4)
        self.assertAlmostEqual(3.685001, self.simpleTriGram.getPerplexity(self.simpleCorpus), 4)

    def test_PerplexityComplex(self):
        self.assertAlmostEqual(1290.97916, self.complexUniGram.getPerplexity(self.testCorpus), 4)
        self.assertAlmostEqual(8331.518540, self.complexBiGram.getPerplexity(self.testCorpus), 4)
        self.assertAlmostEqual(39184.430078, self.complexTriGram.getPerplexity(self.testCorpus), 4)

    def test_CalculateNGramProbabilitiesSimple(self):
        self.assertAlmostEqual(0.116607, self.simpleUniGram.getProbability("<s>"), 4)
        self.assertAlmostEqual(0.149464, self.simpleUniGram.getProbability("mahmut"), 4)
        self.assertAlmostEqual(0.026599, self.simpleUniGram.getProbability("kitabı"), 4)
        self.assertAlmostEqual(0.492147, self.simpleBiGram.getProbability("<s>", "ali"), 4)
        self.assertAlmostEqual(0.030523, self.simpleBiGram.getProbability("ayşe", "ali"), 4)
        self.assertAlmostEqual(0.0625, self.simpleBiGram.getProbability("mahmut", "ali"), 4)
        self.assertAlmostEqual(0.323281, self.simpleBiGram.getProbability("at", "mehmet"), 4)
        self.assertAlmostEqual(0.049190, self.simpleTriGram.getProbability("<s>", "ali", "top"), 4)
        self.assertAlmostEqual(0.043874, self.simpleTriGram.getProbability("ayşe", "kitabı", "at"), 4)
        self.assertAlmostEqual(0.0625, self.simpleTriGram.getProbability("ayşe", "topu", "at"), 4)
        self.assertAlmostEqual(0.0625, self.simpleTriGram.getProbability("mahmut", "evde", "kal"), 4)
        self.assertAlmostEqual(0.261463, self.simpleTriGram.getProbability("ali", "topu", "at"), 4)

    def test_CalculateNGramProbabilitiesComplex(self):
        self.assertAlmostEqual(0.050745, self.complexUniGram.getProbability("<s>"), 4)
        self.assertAlmostEqual(0.000126, self.complexUniGram.getProbability("atatürk"), 4)
        self.assertAlmostEqual(0.000497, self.complexBiGram.getProbability("<s>", "mustafa"), 4)
        self.assertAlmostEqual(0.014000, self.complexBiGram.getProbability("mustafa", "kemal"), 4)
        self.assertAlmostEqual(0.061028, self.complexTriGram.getProbability("<s>", "mustafa", "kemal"), 4)
        self.assertAlmostEqual(0.283532, self.complexTriGram.getProbability("mustafa", "kemal", "atatürk"), 4)


if __name__ == '__main__':
    unittest.main()



import unittest

from NGram.NGram import NGram
from test.CorpusTest import CorpusTest


class SimpleSmoothingTest(CorpusTest, unittest.TestCase):

    simpleUniGram: NGram
    simpleBiGram: NGram
    simpleTriGram: NGram
    complexUniGram: NGram
    complexBiGram: NGram
    complexTriGram: NGram
    simpleCorpus : list
    trainCorpus: list
    testCorpus: list
    validationCorpus: list

    def setUp(self) -> None:
        self.simpleCorpus = [["<s>", "ali", "topu", "at", "mehmet", "ayşeye", "gitti", "</s>"],
                             ["<s>", "ali", "top", "at", "ayşe", "eve", "gitti", "</s>"],
                             ["<s>", "ayşe", "kitabı", "ver", "</s>"],
                             ["<s>", "ali", "topu", "mehmete", "at", "</s>"],
                             ["<s>", "ali", "topu", "at", "mehmet", "ayşeyle", "gitti", "</s>"]]
        self.simpleUniGram = NGram(1, self.simpleCorpus)
        self.simpleBiGram = NGram(2, self.simpleCorpus)
        self.simpleTriGram = NGram(3, self.simpleCorpus)
        self.trainCorpus = self.readCorpus("../train.txt")
        self.complexUniGram = NGram(1, self.trainCorpus)
        self.complexBiGram = NGram(2, self.trainCorpus)
        self.complexTriGram = NGram(3, self.trainCorpus)
        self.testCorpus = self.readCorpus("../test.txt")
        self.validationCorpus = self.readCorpus("../validation.txt")



from Sampling.KFoldCrossValidation import KFoldCrossValidation

from NGram.NGram import NGram
from NGram.TrainedSmoothing import TrainedSmoothing
import math


class AdditiveSmoothing(TrainedSmoothing):

    __delta: float

    def __learnBestDelta(self, nGrams: list, kFoldCrossValidation: KFoldCrossValidation, lowerBound: float) -> float:
        """
        The algorithm tries to optimize the best delta for a given corpus. The algorithm uses perplexity on the
        validation set as the optimization criterion.

        PARAMETERS
        ----------
        nGrams : list
            10 N-Grams learned for different folds of the corpus. nGrams[i] is the N-Gram trained with i'th train fold
            of the corpus.
        kFoldCrossValidation: KFoldCrossValidation
            Cross-validation data used in training and testing the N-grams.
        lowerBound : float
            Initial lower bound for optimizing the best delta.

        RETURNS
        -------
        float
            Best delta optimized with k-fold crossvalidation.
        """
        bestPrevious = -1
        upperBound = 1
        bestDelta = (lowerBound + upperBound) / 2
        numberOfParts = 5
        while True:
            bestPerplexity = 100000000
            value = lowerBound
            while value <= upperBound:
                perplexity = 0
                for i in range(0, 10):
                    nGrams[i].setProbabilityWithPseudoCount(value, nGrams[i].getN())
                    perplexity += nGrams[i].getPerplexity(kFoldCrossValidation.getTestFold(i))
                if perplexity < bestPerplexity:
                    bestPerplexity = perplexity
                    bestDelta = value
                value += (upperBound - lowerBound) / numberOfParts
            lowerBound = self.newLowerBound(bestDelta, lowerBound, upperBound, numberOfParts)
            upperBound = self.newUpperBound(bestDelta, lowerBound, upperBound, numberOfParts)
            if bestPrevious != -1:
                if math.fabs(bestPrevious - bestPerplexity) / bestPerplexity < 0.001:
                    break
            bestPrevious = bestPerplexity
        return bestDelta

    def learnParameters(self, corpus: list, N: int):
        """
        Wrapper function to learn the parameter (delta) in additive smoothing. The function first creates K NGrams
        with the train folds of the corpus. Then optimizes delta with respect to the test folds of the corpus.

        PARAMETERS
        ----------
        corpus : list
            Train corpus used to optimize delta parameter
        N : int
            N in N-Gram.
        """
        K = 10
        nGrams = []
        kFoldCrossValidation = KFoldCrossValidation(corpus, K, 0)
        for i in range(K):
            nGrams.append(NGram(N, kFoldCrossValidation.getTrainFold(i)))
        self.__delta = self.__learnBestDelta(nGrams, kFoldCrossValidation, 0.1)

    def setProbabilities(self, nGram: NGram, level: int):
        """
        Wrapper function to set the N-gram probabilities with additive smoothing.

        PARAMETERS
        ----------
        nGram : NGram
            N-Gram for which the probabilities will be set.
        level : int
            Level for which N-Gram probabilities will be set. Probabilities for different levels of the N-gram can be
            set with this function. If level = 1, N-Gram is treated as UniGram, if level = 2, N-Gram is treated as
            Bigram, etc.
        """
        nGram.setProbabilityWithPseudoCount(self.__delta, level)

    def getDelta(self) -> float:
        """
        Gets the best delta.

        RETURNS
        -------
        float
            learned best delta
        """
        return self.__delta



from setuptools import setup

setup(
    name='NlpToolkit-NGram',
    version='1.0.13',
    packages=['NGram', 'test'],
    url='https://github.com/olcaytaner/NGram-Py',
    license='',
    author='olcaytaner',
    author_email='olcaytaner@isikun.edu.tr',
    description='NGram library',
    install_requires=['NlpToolkit-DataStructure', 'NlpToolkit-Sampling']
)



import unittest

from NGram.InterpolatedSmoothing import InterpolatedSmoothing
from test.SimpleSmoothingTest import SimpleSmoothingTest


class InterpolatedSmoothingTest(SimpleSmoothingTest, unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        interpolatedSmoothing = InterpolatedSmoothing()
        self.complexBiGram.calculateNGramProbabilitiesTrained(self.validationCorpus, interpolatedSmoothing)
        self.complexTriGram.calculateNGramProbabilitiesTrained(self.validationCorpus, interpolatedSmoothing)

    def test_PerplexityComplex(self):
        self.assertAlmostEqual(917.214864, self.complexBiGram.getPerplexity(self.testCorpus), 4)
        self.assertAlmostEqual(3000.451177, self.complexTriGram.getPerplexity(self.testCorpus), 4)

    def test_CalculateNGramProbabilitiesComplex(self):
        self.assertAlmostEqual(0.000418, self.complexBiGram.getProbability("<s>", "mustafa"), 4)
        self.assertAlmostEqual(0.005555, self.complexBiGram.getProbability("mustafa", "kemal"), 4)
        self.assertAlmostEqual(0.014406, self.complexTriGram.getProbability("<s>", "mustafa", "kemal"), 4)
        self.assertAlmostEqual(0.058765, self.complexTriGram.getProbability("mustafa", "kemal", "atatürk"), 4)


if __name__ == '__main__':
    unittest.main()
